# 12 "src/session/strategy_parser.mll"
 
  open Strategy

  exception SyntaxError of string

  let error f =
    Printf.kbprintf
      (fun b ->
        let s = Buffer.contents b in Buffer.clear b; raise (SyntaxError s))
      (Buffer.create 1024)
      f

  type label = {
    mutable defined: int option;
    temporary: int;
  }

  type 'a code = {
    env: Env.env;
    whyconf: Whyconf.config;
    mutable instr: instruction array;
    mutable next: int;
    labels: (string, label) Hashtbl.t; (* label name -> label *)
    mutable temp: int;
  }

  let create_code env conf =
    let h = Hashtbl.create 17 in
    Hashtbl.add h "exit" { defined = Some (-1); temporary = 0 };
    { env = env;
      whyconf = conf;
      instr = Array.make 10 (Igoto 0);
      next = 0;
      temp = 1;
      labels = h; }

  let enlarge_code code =
    let old = code.instr in
    let n = Array.length old in
    code.instr <- Array.make (2 * n) (Igoto 0);
    Array.blit old 0 code.instr 0 n

  let add_instr code i =
    let n = code.next in
    if n = Array.length code.instr then enlarge_code code;
    code.instr.(n) <- i;
    code.next <- n + 1

  let temp code =
    let t = code.temp in code.temp <- t + 1; t

  let define_label code l =
    let n = code.next in
    try
      let lab = Hashtbl.find code.labels l in
      if lab.defined = None then lab.defined <- Some n
      else error "duplicate label %s" l
    with Not_found ->
      let lab =  { defined = Some n; temporary = temp code } in
      Hashtbl.add code.labels l lab

  let find_label code l =
    try
      let lab = Hashtbl.find code.labels l in lab.temporary
    with Not_found ->
      let t = temp code in
      Hashtbl.add code.labels l { defined = None; temporary = t };
      t

  let prover code p =
    try
      let fp = Whyconf.parse_filter_prover p in
      Whyconf.filter_one_prover code.whyconf fp
    with
      | Whyconf.ProverNotFound _ ->
          error "Prover %S not installed or not configured" p
      | Whyconf.ProverAmbiguity _ ->
          error "Prover description %s is ambiguous" p

  let integer msg s =
    try Some (int_of_string s)
    with Failure _ ->
      match s with
      | "%t" -> None
      | "%.t" -> None
      | "%T" -> None
      | "%m" -> None
      | "%s" -> None
      | _ ->
          error "unable to parse %s argument '%s'" msg s

  let real msg s =
    try Some (float_of_string s)
    with Failure _ ->
      match s with
      | "%t" -> None
      | "%.t" -> None
      | "%T" -> None
      | "%S" -> None
      | "%m" -> None
      | "%s" -> None
      | _ ->
          error "unable to parse %s argument '%s'" msg s


  let transform code t =
    try
      ignore (Trans.lookup_transform t code.env)
    with Trans.UnknownTrans _ ->
    try
      ignore (Trans.lookup_transform_l t code.env)
    with Trans.UnknownTrans _->
      error "transformation %S is unknown" t


# 118 "src/session/strategy_parser.ml"
let __ocaml_lex_tables = {
  Lexing.lex_base =
   "\000\000\248\255\249\255\002\000\007\000\009\000\014\000\004\000\
    \016\000\254\255\011\000\042\000\253\255\044\000\053\000\060\000\
    \073\000\080\000\088\000\100\000\117\000\124\000\130\000\141\000\
    \045\000\167\000\194\000\223\000\042\001\005\000\053\001\102\001\
    \002\000\233\000\251\255\132\000\092\001\151\001\117\001\196\000\
    \003\000\026\001\177\001\179\001\184\001\208\001\213\001\215\001\
    \221\001\223\001\251\001\003\002\027\002";
  Lexing.lex_backtrk =
   "\255\255\255\255\255\255\006\000\006\000\006\000\006\000\006\000\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\003\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\004\000\255\255\
    \255\255\004\000\255\255\004\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\005\000";
  Lexing.lex_default =
   "\003\000\000\000\000\000\013\000\013\000\013\000\013\000\010\000\
    \255\255\000\000\010\000\017\000\000\000\013\000\013\000\013\000\
    \013\000\017\000\022\000\013\000\013\000\013\000\022\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\050\000\013\000\013\000\013\000\013\000\013\000\013\000\
    \013\000\013\000\050\000\052\000\052\000";
  Lexing.lex_trans =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\008\000\008\000\255\255\255\255\008\000\009\000\255\255\
    \041\000\041\000\018\000\018\000\041\000\009\000\018\000\011\000\
    \011\000\008\000\008\000\011\000\000\000\008\000\000\000\000\000\
    \008\000\000\000\255\255\007\000\000\000\255\255\000\000\041\000\
    \000\000\018\000\255\255\000\000\255\255\000\000\011\000\000\000\
    \008\000\255\255\000\000\011\000\011\000\255\255\255\255\011\000\
    \000\000\255\255\002\000\000\000\012\000\000\000\255\255\255\255\
    \000\000\012\000\255\255\012\000\000\000\255\255\255\255\000\000\
    \012\000\255\255\011\000\000\000\255\255\255\255\000\000\255\255\
    \000\000\000\000\011\000\011\000\000\000\255\255\011\000\000\000\
    \255\255\255\255\255\255\040\000\255\255\255\255\000\000\255\255\
    \000\000\018\000\018\000\005\000\255\255\018\000\012\000\006\000\
    \000\000\011\000\019\000\000\000\255\255\255\255\255\255\012\000\
    \255\255\255\255\035\000\255\255\004\000\034\000\012\000\039\000\
    \018\000\042\000\000\000\255\255\000\000\014\000\255\255\255\255\
    \000\000\039\000\255\255\012\000\255\255\018\000\018\000\255\255\
    \000\000\018\000\255\255\023\000\023\000\031\000\031\000\023\000\
    \000\000\031\000\255\255\000\000\000\000\255\255\023\000\023\000\
    \255\255\000\000\023\000\000\000\018\000\000\000\012\000\255\255\
    \000\000\039\000\023\000\000\000\031\000\255\255\000\000\000\000\
    \000\000\015\000\000\000\016\000\000\000\023\000\000\000\012\000\
    \028\000\028\000\024\000\000\000\028\000\000\000\012\000\000\000\
    \027\000\000\000\027\000\025\000\255\255\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\026\000\026\000\028\000\
    \000\000\000\000\000\000\028\000\028\000\028\000\028\000\028\000\
    \020\000\028\000\000\000\000\000\000\000\000\000\000\000\025\000\
    \025\000\025\000\025\000\025\000\025\000\025\000\025\000\025\000\
    \025\000\021\000\028\000\000\000\028\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\036\000\000\000\000\000\000\000\
    \025\000\000\000\026\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\000\000\000\000\000\000\000\000\
    \001\000\000\000\255\255\000\000\009\000\000\000\000\000\255\255\
    \000\000\255\255\000\000\009\000\036\000\025\000\255\255\027\000\
    \027\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\041\000\041\000\000\000\000\000\041\000\
    \000\000\000\000\255\255\000\000\255\255\000\000\000\000\000\000\
    \000\000\000\000\000\000\028\000\028\000\255\255\000\000\028\000\
    \000\000\000\000\041\000\000\000\255\255\255\255\031\000\031\000\
    \000\000\000\000\031\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\255\255\028\000\000\000\000\000\000\000\000\000\029\000\
    \255\255\000\000\000\000\000\000\255\255\031\000\000\000\000\000\
    \255\255\000\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\255\255\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\031\000\
    \031\000\000\000\000\000\031\000\000\000\255\255\000\000\000\000\
    \000\000\000\000\000\000\000\000\255\255\000\000\000\000\000\000\
    \000\000\000\000\255\255\000\000\000\000\000\000\031\000\038\000\
    \000\000\038\000\000\000\032\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \028\000\028\000\000\000\000\000\028\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\028\000\
    \000\000\000\000\255\255\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\000\000\000\000\255\255\000\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\255\255\000\000\255\255\255\255\000\000\255\255\000\000\
    \255\255\255\255\255\255\255\255\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\000\000\255\255\000\000\255\255\255\255\
    \041\000\041\000\255\255\012\000\041\000\012\000\000\000\000\000\
    \255\255\000\000\012\000\255\255\000\000\255\255\000\000\255\255\
    \255\255\000\000\255\255\000\000\000\000\255\255\000\000\041\000\
    \255\255\000\000\255\255\000\000\051\000\051\000\000\000\000\000\
    \051\000\000\000\012\000\000\000\051\000\051\000\000\000\012\000\
    \051\000\012\000\043\000\000\000\000\000\000\000\000\000\012\000\
    \000\000\012\000\255\255\051\000\000\000\000\000\255\255\000\000\
    \000\000\044\000\000\000\051\000\255\255\255\255\255\255\000\000\
    \255\255\000\000\000\000\045\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\255\255\046\000\000\000\
    \000\000\000\000\000\000\255\255\000\000\255\255\255\255\000\000\
    \000\000\000\000\000\000\000\000\047\000\000\000\000\000\000\000\
    \000\000\048\000\049\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\255\255\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\255\255\000\000\255\255\000\000\000\000\000\000\000\000\
    \255\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \255\255\000\000\000\000\000\000\000\000\255\255\000\000\255\255\
    \000\000\000\000\000\000\000\000\000\000\255\255\000\000\255\255\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\255\255\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\255\255\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\255\255";
  Lexing.lex_check =
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\003\000\003\000\000\000\007\000\003\000\
    \004\000\004\000\005\000\005\000\004\000\010\000\005\000\006\000\
    \006\000\008\000\008\000\006\000\255\255\008\000\255\255\255\255\
    \000\000\255\255\003\000\000\000\255\255\003\000\255\255\004\000\
    \255\255\005\000\004\000\255\255\005\000\255\255\006\000\255\255\
    \008\000\006\000\255\255\011\000\011\000\013\000\013\000\011\000\
    \255\255\013\000\000\000\255\255\003\000\255\255\014\000\014\000\
    \255\255\004\000\014\000\005\000\255\255\015\000\015\000\255\255\
    \006\000\015\000\011\000\255\255\013\000\011\000\255\255\013\000\
    \255\255\255\255\016\000\016\000\255\255\014\000\016\000\255\255\
    \014\000\017\000\017\000\024\000\015\000\017\000\255\255\015\000\
    \255\255\018\000\018\000\000\000\011\000\018\000\013\000\000\000\
    \255\255\016\000\005\000\255\255\016\000\019\000\019\000\014\000\
    \017\000\019\000\029\000\017\000\000\000\032\000\015\000\040\000\
    \018\000\004\000\255\255\018\000\255\255\006\000\020\000\020\000\
    \255\255\024\000\020\000\016\000\019\000\021\000\021\000\019\000\
    \255\255\021\000\017\000\022\000\022\000\035\000\035\000\022\000\
    \255\255\035\000\018\000\255\255\255\255\020\000\023\000\023\000\
    \020\000\255\255\023\000\255\255\021\000\255\255\019\000\021\000\
    \255\255\024\000\022\000\255\255\035\000\022\000\255\255\255\255\
    \255\255\014\000\255\255\015\000\255\255\023\000\255\255\020\000\
    \025\000\025\000\023\000\255\255\025\000\255\255\021\000\255\255\
    \023\000\255\255\023\000\023\000\022\000\023\000\023\000\023\000\
    \023\000\023\000\023\000\023\000\023\000\023\000\023\000\025\000\
    \255\255\255\255\255\255\026\000\026\000\039\000\039\000\026\000\
    \019\000\039\000\255\255\255\255\255\255\255\255\255\255\025\000\
    \025\000\025\000\025\000\025\000\025\000\025\000\025\000\025\000\
    \025\000\020\000\026\000\255\255\039\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\025\000\255\255\255\255\255\255\
    \026\000\255\255\026\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\255\255\255\255\255\255\255\255\
    \000\000\255\255\003\000\255\255\007\000\255\255\255\255\004\000\
    \255\255\005\000\255\255\010\000\025\000\027\000\006\000\027\000\
    \027\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\041\000\041\000\255\255\255\255\041\000\
    \255\255\255\255\011\000\255\255\013\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\028\000\028\000\014\000\255\255\028\000\
    \255\255\255\255\041\000\255\255\015\000\041\000\030\000\030\000\
    \255\255\255\255\030\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\016\000\028\000\255\255\255\255\255\255\255\255\028\000\
    \017\000\255\255\255\255\255\255\041\000\030\000\255\255\255\255\
    \018\000\255\255\028\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\019\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\031\000\
    \031\000\255\255\255\255\031\000\255\255\020\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\021\000\255\255\255\255\255\255\
    \255\255\255\255\022\000\255\255\255\255\255\255\031\000\036\000\
    \255\255\036\000\255\255\031\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\036\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \037\000\037\000\255\255\255\255\037\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\037\000\
    \255\255\255\255\042\000\042\000\043\000\043\000\042\000\255\255\
    \043\000\044\000\044\000\255\255\255\255\044\000\255\255\037\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\042\000\255\255\043\000\042\000\255\255\043\000\255\255\
    \044\000\045\000\045\000\044\000\255\255\045\000\046\000\046\000\
    \047\000\047\000\046\000\255\255\047\000\255\255\048\000\048\000\
    \049\000\049\000\048\000\042\000\049\000\043\000\255\255\255\255\
    \045\000\255\255\044\000\045\000\255\255\046\000\255\255\047\000\
    \046\000\255\255\047\000\255\255\255\255\048\000\255\255\049\000\
    \048\000\255\255\049\000\255\255\050\000\050\000\255\255\255\255\
    \050\000\255\255\045\000\255\255\051\000\051\000\255\255\046\000\
    \051\000\047\000\042\000\255\255\255\255\255\255\255\255\048\000\
    \255\255\049\000\041\000\050\000\255\255\255\255\050\000\255\255\
    \255\255\043\000\255\255\051\000\052\000\052\000\051\000\255\255\
    \052\000\255\255\255\255\044\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\050\000\045\000\255\255\
    \255\255\255\255\255\255\052\000\255\255\051\000\052\000\255\255\
    \255\255\255\255\255\255\255\255\046\000\255\255\255\255\255\255\
    \255\255\047\000\048\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\052\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\042\000\255\255\043\000\255\255\255\255\255\255\255\255\
    \044\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \045\000\255\255\255\255\255\255\255\255\046\000\255\255\047\000\
    \255\255\255\255\255\255\255\255\255\255\048\000\255\255\049\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\050\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\051\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\052\000";
  Lexing.lex_base_code =
   "\000\000\000\000\000\000\000\000\000\000\002\000\007\000\000\000\
    \000\000\000\000\000\000\009\000\000\000\000\000\000\000\000\000\
    \014\000\000\000\016\000\000\000\000\000\027\000\034\000\045\000\
    \000\000\055\000\104\000\003\000\114\000\006\000\163\000\173\000\
    \003\000\183\000\049\000\043\000\193\000\243\000\209\000\112\000\
    \003\000\013\001\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\119\000\015\001\165\000\000\000";
  Lexing.lex_backtrk_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\010\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\028\000\000\000\
    \000\000\049\000\000\000\028\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\072\000";
  Lexing.lex_default_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\013\000\000\000\000\000\000\000\013\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\066\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\066\000\000\000\000\000";
  Lexing.lex_trans_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\001\000\001\000\004\000\004\000\001\000\000\000\004\000\
    \007\000\007\000\007\000\007\000\007\000\000\000\007\000\007\000\
    \007\000\004\000\004\000\007\000\000\000\004\000\000\000\000\000\
    \001\000\000\000\004\000\000\000\004\000\004\000\000\000\007\000\
    \004\000\007\000\000\000\016\000\016\000\000\000\007\000\016\000\
    \004\000\019\000\000\000\000\000\043\000\043\000\016\000\016\000\
    \043\000\000\000\016\000\004\000\000\000\000\000\000\000\000\000\
    \022\000\022\000\016\000\000\000\022\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\043\000\000\000\016\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\019\000\000\000\000\000\022\000\
    \000\000\000\000\000\000\019\000\000\000\019\000\019\000\019\000\
    \019\000\019\000\019\000\019\000\019\000\019\000\019\000\019\000\
    \019\000\019\000\019\000\019\000\019\000\019\000\019\000\019\000\
    \019\000\022\000\022\000\025\000\019\000\022\000\046\000\019\000\
    \000\000\022\000\022\000\022\000\022\000\022\000\000\000\022\000\
    \001\000\001\000\000\000\000\000\001\000\000\000\000\000\000\000\
    \022\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \022\000\000\000\022\000\000\000\000\000\000\000\019\000\001\000\
    \019\000\019\000\019\000\019\000\019\000\019\000\019\000\019\000\
    \019\000\019\000\025\000\025\000\025\000\025\000\025\000\025\000\
    \025\000\025\000\025\000\025\000\043\000\043\000\069\000\069\000\
    \043\000\000\000\069\000\000\000\000\000\000\000\043\000\043\000\
    \000\000\000\000\043\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\043\000\000\000\069\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\043\000\000\000\000\000\
    \000\000\000\000\000\000\025\000\025\000\025\000\025\000\025\000\
    \025\000\025\000\025\000\025\000\025\000\046\000\046\000\046\000\
    \046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\
    \046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\
    \046\000\019\000\019\000\019\000\019\000\019\000\019\000\019\000\
    \019\000\019\000\019\000\000\000\022\000\022\000\000\000\000\000\
    \022\000\019\000\019\000\019\000\019\000\019\000\019\000\019\000\
    \019\000\019\000\019\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\022\000\000\000\000\000\001\000\001\000\
    \069\000\069\000\001\000\000\000\069\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\019\000\019\000\019\000\019\000\019\000\
    \019\000\019\000\019\000\019\000\019\000\001\000\000\000\069\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    ";
  Lexing.lex_check_code =
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\004\000\004\000\005\000\005\000\004\000\255\255\005\000\
    \006\000\006\000\011\000\011\000\006\000\255\255\011\000\016\000\
    \016\000\018\000\018\000\016\000\255\255\018\000\255\255\255\255\
    \004\000\255\255\005\000\255\255\021\000\021\000\255\255\006\000\
    \021\000\011\000\255\255\022\000\022\000\255\255\016\000\022\000\
    \018\000\027\000\255\255\018\000\035\000\035\000\023\000\023\000\
    \035\000\255\255\023\000\021\000\255\255\255\255\255\255\255\255\
    \025\000\025\000\022\000\255\255\025\000\022\000\255\255\255\255\
    \255\255\255\255\018\000\035\000\255\255\023\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\024\000\255\255\255\255\025\000\
    \255\255\255\255\255\255\023\000\022\000\023\000\023\000\023\000\
    \023\000\023\000\023\000\023\000\023\000\023\000\023\000\025\000\
    \025\000\025\000\025\000\025\000\025\000\025\000\025\000\025\000\
    \025\000\026\000\026\000\029\000\024\000\026\000\032\000\040\000\
    \255\255\039\000\039\000\028\000\028\000\039\000\255\255\028\000\
    \049\000\049\000\255\255\255\255\049\000\255\255\255\255\255\255\
    \026\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \039\000\255\255\028\000\255\255\255\255\255\255\026\000\049\000\
    \026\000\026\000\026\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\030\000\030\000\051\000\051\000\
    \030\000\255\255\051\000\255\255\255\255\255\255\031\000\031\000\
    \255\255\255\255\031\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\030\000\255\255\051\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\031\000\255\255\255\255\
    \255\255\255\255\255\255\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\255\255\037\000\037\000\255\255\255\255\
    \037\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\255\255\255\255\255\255\255\255\255\255\
    \018\000\255\255\255\255\037\000\255\255\255\255\041\000\041\000\
    \050\000\050\000\041\000\255\255\050\000\255\255\255\255\255\255\
    \255\255\255\255\022\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\041\000\255\255\050\000\
    \041\000\255\255\050\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\041\000\
    \255\255\050\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\041\000\255\255\050\000\
    ";
  Lexing.lex_code =
   "\255\017\255\255\009\255\255\008\255\255\000\008\255\010\255\255\
    \011\255\255\012\255\255\013\255\255\014\255\255\007\255\002\011\
    \003\012\000\009\001\010\004\013\005\014\255\015\255\255\016\255\
    \255\002\011\003\012\007\015\006\016\000\009\001\010\004\013\005\
    \014\255\018\255\255\019\255\255\000\017\001\018\002\019\255";
}

let rec scan code lexbuf =
  lexbuf.Lexing.lex_mem <- Array.make 20 (-1); __ocaml_lex_scan_rec code lexbuf 0
and __ocaml_lex_scan_rec code lexbuf __ocaml_lex_state =
  match Lexing.new_engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 144 "src/session/strategy_parser.mll"
      ( scan code lexbuf )
# 521 "src/session/strategy_parser.ml"

  | 1 ->
# 146 "src/session/strategy_parser.mll"
      ( scan code lexbuf )
# 526 "src/session/strategy_parser.ml"

  | 2 ->
let
# 147 "src/session/strategy_parser.mll"
             id
# 532 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos (lexbuf.Lexing.lex_curr_pos + -1) in
# 148 "src/session/strategy_parser.mll"
      ( define_label code id;
        scan code lexbuf )
# 537 "src/session/strategy_parser.ml"

  | 3 ->
let
# 150 "src/session/strategy_parser.mll"
                          id
# 543 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_curr_pos in
# 151 "src/session/strategy_parser.mll"
      ( add_instr code (Igoto (find_label code id));
        scan code lexbuf )
# 548 "src/session/strategy_parser.ml"

  | 4 ->
let
# 153 "src/session/strategy_parser.mll"
                          p
# 554 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_mem.(1)
and
# 153 "src/session/strategy_parser.mll"
                                                  t
# 559 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(2) lexbuf.Lexing.lex_mem.(3)
and
# 153 "src/session/strategy_parser.mll"
                                                                         m
# 564 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(4) lexbuf.Lexing.lex_mem.(5)
and
# 154 "src/session/strategy_parser.mll"
                               s
# 569 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme_opt lexbuf lexbuf.Lexing.lex_mem.(7) lexbuf.Lexing.lex_mem.(6) in
# 155 "src/session/strategy_parser.mll"
      ( let p = prover code p in
        let t = real "timelimit" t in
        if t <> None && Opt.get t <= 0.0 then
          error "timelimit %f is invalid" (Opt.get t);
        let m = integer "memlimit" m in
        if m <> None && Opt.get m <= 0 then
          error "memlimit %d is invalid" (Opt.get m);
        let s = integer "steplimit" (Opt.get_def "%s" s) in
        if s <> None && Opt.get s < 0 then
          error "steplimit %d is invalid" (Opt.get s);
        add_instr code (Icall_prover (p.Whyconf.prover, t, m, s));
        scan code lexbuf )
# 584 "src/session/strategy_parser.ml"

  | 5 ->
let
# 167 "src/session/strategy_parser.mll"
                               t
# 590 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_mem.(1)
and
# 167 "src/session/strategy_parser.mll"
                                                   l
# 595 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(2) lexbuf.Lexing.lex_curr_pos in
# 168 "src/session/strategy_parser.mll"
      ( transform code t;
        add_instr code (Itransform (t, find_label code l));
        scan code lexbuf )
# 601 "src/session/strategy_parser.ml"

  | 6 ->
let
# 171 "src/session/strategy_parser.mll"
         c
# 607 "src/session/strategy_parser.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 172 "src/session/strategy_parser.mll"
      ( let i = Lexing.lexeme_start lexbuf in
        error "syntax error on character '%c' at position %d" c i )
# 612 "src/session/strategy_parser.ml"

  | 7 ->
# 175 "src/session/strategy_parser.mll"
      ( () )
# 617 "src/session/strategy_parser.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_scan_rec code lexbuf __ocaml_lex_state

;;

# 177 "src/session/strategy_parser.mll"
 

  let parse env conf s =
    let code = create_code env conf in
    scan code (Lexing.from_string s);
    let label = Array.make code.temp 0 in
    let fill name lab = match lab.defined with
      | None -> error "label '%s' is undefined" name
      | Some n -> label.(lab.temporary) <- n in
    Hashtbl.iter fill code.labels;
    let solve = function
      | Icall_prover _ as i -> i
      | Itransform (t, n) -> Itransform (t, label.(n))
      | Igoto n -> Igoto label.(n) in
    Array.map solve (Array.sub code.instr 0 code.next)


# 642 "src/session/strategy_parser.ml"
